(use-modules ((srfi srfi-1)  :select ((for-each . for-each-1) circular-list)))
(use-modules ((srfi srfi-13) :select
              (string-append/shared
               substring/shared string-prefix? string-suffix?)))
(use-modules ((ice-9 q)      :select (make-q enq!)))
(use-modules ((ice-9 rdelim) :select (read-line)))

(define (write-file port-or-path text)
  "Write the given TEXT to the given PORT-OR-PATH, and close the port."
  (let ((port (if (port? port-or-path)
                port-or-path
                (open port-or-path (logior O_WRONLY O_CREAT O_EXCL)))))
    (display text port)
    (close-port port)))

(define (slurp port-or-path)
  "Read and return all text from the given PORT-OR-PATH, and close the port."
  (let ((port (if (port? port-or-path)
                port-or-path
                (open port-or-path O_RDONLY))))
    (let loop ((buffer ""))
      (if (eof-object? (peek-char port))
        (begin
          (close-port port)
          buffer)
        (loop (string-append/shared buffer (drain-input port)))))))

(define (lines-until-empty port-or-path . affixes)
  "\
Return a list of lines read from the given PORT-OR-PATH, up to the first line
that is empty or does not begin with PREFIX and end with SUFFIX.  PREFIX and
SUFFIX are removed from each line."
  (let* ((port (if (port? port-or-path)
                 port-or-path
                 (open port-or-path O_RDONLY)))
         (lines (make-q))
         (prefix (if (null? affixes) "" (car  affixes)))
         (suffix (if (null? affixes) "" (cadr affixes)))
         (plen (string-length prefix))
         (slen (string-length suffix))
         (len (+ plen slen)))
    (let loop ()
      (let* ((line (read-line port)))
        (if (or (eof-object? line)
                (string-null? line)
                (< (string-length line) len)
                (not (string-prefix? prefix line))
                (not (string-suffix? suffix line)))
          (begin
            (close-port port)
            (car lines))
          (begin
            (enq! lines (substring/shared line plen
                                          (- (string-length line) slen)))
            (loop)))))))

(define (cat ports-or-paths . output-port)
  "\
Write the contents of each of the PORTS-OR-PATHS to the given OUTPUT-PORT, or
the current output port by default, and close the input ports."
  (let ((output-port
         (if (null? output-port) (current-output-port) (car output-port))))
    (for-each-1 cat-port-or-path ports-or-paths (circular-list output-port))))
(define (cat-port-or-path port-or-path output-port)
  (cat-port (if (port? port-or-path)
              port-or-path
              (open port-or-path O_RDONLY))
            output-port))
(define (cat-port input-port output-port)
  (while (not (eof-object? (peek-char input-port)))
    (display (drain-input input-port) output-port))
  (close-port input-port))
